import streamlit as st
from PIL import Image

from rqle_ai_langchain_util.utils import gui_util
from rqle_ai_langchain_util.apps.clarity_genie.clarity_genie import ClarityGenie

APPLICATION_LOGO_PATH = 'img/claritygenie_logo.jpg'
APPLICATION_LOGO = Image.open(APPLICATION_LOGO_PATH)
APPLICATION_NAME = 'ClarityGenie'
APPLICATION_DESCRIPTION = f"""
    **{APPLICATION_NAME}**:copyright: is your one-stop shop for crafting powerful prompts fueled by cutting-edge 
    Generative AI. It aims to (i) refine your prompts for ultimate clarity, and (ii) add context to your prompts.
"""

# set up the high-level configuration of the page
st.set_page_config(page_title=APPLICATION_NAME, page_icon=APPLICATION_LOGO,
                   layout='wide', initial_sidebar_state='collapsed')

# information about / help component of the application
st.sidebar = gui_utils.set_page_sidebar(application_description=APPLICATION_DESCRIPTION)

# set the logo / title of the application
st.markdown(body=gui_utils.set_page_header(application_name=APPLICATION_NAME, application_logo_path=APPLICATION_LOGO_PATH),
            unsafe_allow_html=True)

# main body of the application
input, output = st.columns([1.5, 2.5])
output.subheader('Clarified Prompt', divider='rainbow')

with input:
    with st.form(key='prompt_input'):
        st.subheader('Paste Prompt', divider='rainbow',help='Please paste a prompt to be clarified.')
        prompt = st.text_area(label='Prompt',key='prompt', height=400, label_visibility='hidden')
        # form submission button
        submit = st.form_submit_button(label='Clarify', type='primary')

        # execute logic if submit button is clicked
        if submit:
            if not prompt or prompt == '':
                st.error('Please enter a prompt to be clarified.')
            else:
                with output:
                    with st.spinner('Clarifying prompt...'):
                        clarity_genie = ClarityGenie(config_folder='clarity_genie')
                        clarified_prompt = clarity_genie.invoke_chain(prompt)
                        st.write(clarified_prompt['text'])

# information about the footer of the application (including logo and years)
st.markdown(body=gui_utils.set_page_footer(), unsafe_allow_html=True)



import base64
import json
import os
from pathlib import Path


def file_exists(file_dir: str, file_name: str) -> bool:
    """
    :param file_dir: Directory in which JSON file is stored
    :param file_name: Name of the file
    :return: True if the file exists, False otherwise
    """
    return os.path.exists(os.path.join(file_dir, file_name))


def read_file(file_dir: str, file_name: str) -> str:
    """
    :param file_dir: Directory in which JSON file is stored
    :param file_name: Name of the file
    :return: String representation of the content in the file
    """
    input_file = Path(os.path.join(file_dir, file_name))

    # test whether the file is a directory
    if input_file.is_dir():
        raise IsADirectoryError(f'{input_file} is not a file')

    try:
        return input_file.read_text(encoding='utf-8')

    except FileNotFoundError:
        raise FileNotFoundError(f'{file_name} not found in {file_dir}')
    except Exception as e:
        raise IOError(f'Error reading {file_name}: {e}')


def read_json_file(file_dir: str, file_name: str) -> dict:
    """
    :param file_dir: Directory in which JSON file is stored
    :param file_name: Name of the JSON file
    :return: JSON  representation of the content in the JSON file
    """
    input_file = Path(os.path.join(file_dir, file_name))
    # test whether the file is a directory
    if input_file.is_dir():
        raise IsADirectoryError(f'{input_file} is not a JSON file')
    # test whether the file has the extension .json
    if not file_name.endswith('.json'):
        raise TypeError(f'{file_name} is not a JSON file')
    try:
        with input_file.open('r') as f:
            data = json.load(f)

        return data
    except FileNotFoundError:
        raise FileNotFoundError(f'{file_name} not found in {file_dir}')
    except json.JSONDecodeError as jde:
        raise json.JSONDecodeError(msg=f'{file_name} is not a valid JSON file', doc=jde.doc, pos=jde.pos)
    except Exception as e:
        raise IOError(f'Error reading {file_name}: {e}')


def read_image(image_path: str) -> base64:
    """
    Reads an image from a file and returns its base64-encoded representation
    :param image_path: the path to the image on the local file system
    :return: a base64-encoded string representation of the image
    """
    if os.path.isdir(image_path):
        raise IsADirectoryError(f'{image_path} is not a file')

    try:
        with open(image_path, 'rb') as f:
            image = f.read()
            return base64.b64encode(image).decode('utf-8')
    except FileNotFoundError:
        raise FileNotFoundError(f'{image_path} not found')
    except Exception as e:
        raise IOError(f'Error reading {image_path}: {e}')



from rqle_ai_langchain_util.llms.adapters.llm_adapters import LLMAdapter
from rqle_ai_langchain_util.llms.adapters import (aws_bedrock_adapter, azure_openai_adapter, google_gemini_adapter,
                                                  ollama_adapter, oci_ai_adapter)
from rqle_ai_langchain_util.prompts.prompt_config import PromptConfig
from rqle_ai_langchain_util.prompts.prompt_template import PromptTemplate
from rqle_ai_langchain_util.prompts.prompt_example import PromptExample
from rqle_ai_langchain_util.settings import PROMPT_CONFIG_FOLDER
from rqle_ai_langchain_util.utils.file_util import file_exists, read_file


class LLMMediator:

    def info(self) -> str:
        """
        :return: the name of the class
        """
        return self.__class__.__name__

    def __init__(self, llm_adapter: LLMAdapter, prompt_config_name: str):
        """
        :param llm_adapter: Adapter for the execution of the LLM
        :param prompt_config_name: Name of the prompt
        """
        self._llm_adapter = llm_adapter
        self._prompt_name = prompt_config_name
        # populate the prompt config
        self._prompt_config = PromptConfig.from_json(f'{PROMPT_CONFIG_FOLDER}/{prompt_config_name}')
        # populate the prompt template
        if file_exists(f'{PROMPT_CONFIG_FOLDER}/{prompt_config_name}', 'prompt.json'):
            self._prompt_template = PromptTemplate.from_json(f'{PROMPT_CONFIG_FOLDER}/{prompt_config_name}')
        else:
            self._prompt_template = PromptTemplate.from_text(read_file(file_dir=f'{PROMPT_CONFIG_FOLDER}/{prompt_config_name}',
                                                                       file_name='prompt.txt'))
        # populate the prompt example
        if file_exists(f'{PROMPT_CONFIG_FOLDER}/{prompt_config_name}', 'example.json'):
            self._prompt_example = PromptExample.from_json(f'{PROMPT_CONFIG_FOLDER}/{prompt_config_name}')
        self._model = _load_model(llm_adapter, self._prompt_config)

    @property
    def prompt_name(self) -> str:
        """
        :return: Prompt name
        """
        return self._prompt_name

    @property
    def prompt_config(self) -> PromptConfig:
        """
        :return: Prompt configuration
        """
        return self._prompt_config

    @property
    def prompt_template(self) -> PromptTemplate:
        """
        :return: Prompt template
        """
        return self._prompt_template

    @property
    def prompt_example(self) -> PromptExample:
        """
        :return: Prompt example
        """
        return self._prompt_example

    @property
    def model(self):
        """
        :return: LLM model
        """
        return self._model


def _load_model(llm_adapter: LLMAdapter, llm_config: PromptConfig):
    """
    :param llm_adapter: Adapter for the execution of the LLM
    :param llm_config: Configuration for the LLM
    :return: Configured LLM
    """
    llm = None
    if llm_adapter == LLMAdapter.AWS_BEDROCK:
        llm = aws_bedrock_adapter.load_aws_bedrock_from_prompt_config(llm_config)
    elif llm_adapter == LLMAdapter.AZURE_OPENAI:
        llm = azure_openai_adapter.load_azure_openai_from_prompt_config(llm_config)
    elif llm_adapter == LLMAdapter.GOOGLE_GEMINI:
        llm = google_gemini_adapter.load_google_gemini_from_prompt_config(llm_config)
    elif llm_adapter == LLMAdapter.OLLAMA_AI:
        llm = ollama_adapter.load_ollama_from_prompt_config(llm_config)
    elif llm_adapter == LLMAdapter.OCI_AI:
        llm = oci_ai_adapter.load_oci_ai_llm_from_prompt_config(llm_config)
    else:
        raise NotImplementedError(f'LLM adapter {llm_adapter} not supported')
    return llm



import plotly.express as px
import plotly.graph_objects as go
import streamlit as st

from rqle_ai_langchain_util.utils.file_util import read_image

COMPANY_LOGO_PATH = 'img/rqle_ai_logo_alt.jpeg'
COMPANY_NAME = 'RQle.AI'


def set_page_sidebar(application_description: str, background_color: str = '#C0BAAC'):
    """
    Set up the sidebar of the application
    :param application_description: the description of the application
    :param background_color: the background color of the sidebar
    :return: a sidebar to be integrated in a streamlit app
    """
    st.markdown(f"""
        <style>
            [data-testid=stSidebar] {{
                background-color: {background_color};
            }}
        </style>
        """, unsafe_allow_html=True)
    st.sidebar.markdown(body=f"""<img src="data:image/jpeg;base64,{read_image(COMPANY_LOGO_PATH)}" 
                                      alt="{COMPANY_LOGO_PATH}" width="100" height="100"/>""",
                        unsafe_allow_html=True)
    st.sidebar.header(body='About')
    st.sidebar.markdown(body=f"{application_description}")

    st.sidebar.header(body=COMPANY_NAME)
    st.sidebar.markdown(body=f"""
     **{COMPANY_NAME}** believes in the transformative potential of Generative AI. More specifically, it focuses on 
     showcasing real-world applications of how Generative AI can empower individuals and organizations worldwide in 
     addressing their customers' "*job to be done*" problems and create value for them.
    """)

    st.sidebar.header(body='Resources')
    st.sidebar.markdown(body="""
    - [GitHub Repository](https://github.com/RQledotai/rqle-ai-langchain-util)
    - [LinkedIn Profile](https://www.linkedin.com/company/102641077)
    - **Author**: [RQle.ai](mailto:rqledotai@gmail.com)
    """)

    st.sidebar.header(body='Disclaimer')
    st.sidebar.markdown(body="""
    * The development of this application has been aided by Generative AI. More specifically, 
    
    cthrough a combination of 
      [codeium](https://codeium.com/) and either a premium or an open-source large language models to suggest coding 
      best practices.
    * The content and / or algorithms generated by this application are not intended to be final. In other words, the 
      human user has the responsibility to critically evaluate and revise the output to ensure its accuracy, 
      appropriateness, and ethical implications.
    """)

    return st.sidebar


def set_page_header(application_name: str, application_logo_path: str, background_color: str = '#C0BAAC',
                    text_color: str = '#000000') -> str:
    """
    Set up the header of the application
    :param application_name: the name of the application
    :param application_logo_path: the path to the logo of the application
    :param background_color: the background color of the header
    :param text_color: the text color of the header
    :return: Markdown string for the header of the application
    """
    return f"""
     <style>
        header {{
            background-color: {background_color};
        }}
        h1 {{
            color: {text_color};
        }}
     </style>
     <header>
        <h1 align="center">
            <img src="data:image/jpeg;base64,{read_image(application_logo_path)}" alt="{application_logo_path}" 
                 width="200" height="200"/> - 
            {application_name}&copy;
        </h1>
        <hr/>
    </header>
    """


def set_page_footer(years: str = '2024') -> str:
    """
    Set up the footer of the application
    :param years: the years of the copyright
    :return: Markdown string for the footer of the application
    """
    return f"""
     <style>
        footer {{
            background-color: #C0BAAC;
        }}
     </style>
     <footer>
        <hr/>
        <h5 align="center">
            {years} - All rights reserved.
            <img src="data:image/jpeg;base64,{read_image(COMPANY_LOGO_PATH)}" alt="{COMPANY_LOGO_PATH}" 
                 width="100" height="100"/>
        </h5>
    </footer>
    """


def make_donut_graph(expected: int, actual: int):
    """
    Create a donut chart comparing expected vs actual.
    :param expected: the expected value
    :param actual: the actual / estimated value
    :return: a donut graph showing the comparison
    """
    # determine which color palette relate to different donut charts
    green_color_palette = ['#27AE60', '#12783D']
    orange_color_palette = ['#F39C12', '#875A12']
    red_color_palette = ['#E74C3C', '#781F16']

    # calculate the deviation between actual vs expected & decide colors to be used
    deviation_percentage = (actual / expected) * 100
    if (deviation_percentage >= 65):
        chart_color = green_color_palette
    elif 35 < deviation_percentage < 65:
        chart_color = orange_color_palette
    else:
        chart_color = red_color_palette

    labels = ['Expected', 'Actual']
    percents = [100, deviation_percentage]

    # create the donut chart
    fig = go.Figure()
    fig.add_trace(go.Pie(labels=labels, values=percents, rotation=150, hole=0.3, marker_colors=chart_color))
    fig.add_trace(go.Pie(labels=labels, values=percents, rotation=150, hole=0.3, marker_colors=chart_color))
    fig.update_traces()
    fig.update_traces(textinfo='none', title=f'{actual} min', hoverinfo='none')
    fig.update(layout_showlegend=False)

    return fig







from dotenv import load_dotenv

from langchain.chains import LLMChain
from langchain_core.prompts import PromptTemplate

from rqle_ai_langchain_util import settings
from rqle_ai_langchain_util.llms.adapters.llm_adapters import LLMAdapter
from rqle_ai_langchain_util.llms.llm_mediator import LLMMediator

# logging configuration
from time import time
import logging
from logging.config import dictConfig
dictConfig(settings.LOG_CONFIG)
logger = logging.getLogger(__name__)

load_dotenv()


class ClarityGenie:

    def info(self):
        """
        :return: the name of the class
        """
        return self.__class__.__name__

    def __init__(self, config_folder: str):
        """
        Initialise the ClarityGenie class
        :param config_folder: the folder with the prompt configuration
        """
        # configure the LLM to be executed
        self.config_folder = config_folder
        self.llm_mediator = LLMMediator(LLMAdapter.OLLAMA_AI, self.config_folder)

    def load_chain(self):
        """
        :return: a configured LLM to be executed
        """
        try:
            # configure the prompt
            prompt = PromptTemplate(template=self.llm_mediator.prompt_template.prompt,
                                    input_variables=['prompt'])
            logger.debug(f'Generated prompt: {prompt}')

            # return the LLM
            return LLMChain(llm=self.llm_mediator.model, prompt=prompt)
        except Exception as e:
            logger.error(f'Error loading chain: {self.config_folder}\n{e}', exc_info=True)
            raise e

    def invoke_chain(self, input_prompt: str):
        """
        :param input_prompt: the input prompt to be revised
        :return: the output from executing the LLM chain
        """
        try:
            # load the chain to be executed
            chain = self.load_chain()
            # execute the chain
            output = chain.invoke({'prompt': input_prompt})
            logger.debug(f'Output from ClarityGenie: {output}')
            return output
        except Exception as e:
            logger.error(f'Error executing chain: {self.config_folder}\n{e}', exc_info=True)
            raise e


if __name__ == '__main__':
    start = time()
    try:
        # instantiate and the ClarityGenie
        cg = ClarityGenie('clarity_genie')
        output = cg.invoke_chain("""
        Given a job description in triple backticks for {job_title} at {company}, your goal is to re-write the user's past experience delimited by <<<>>> to match the job-description.

The revised resume should include:
- two to three sentences that shine a line on the most relevant / impressive achievements based on the past 10 years
- rewrite past experience with 3-5 bullet per role to make it more actionable
- showcase my 10 most relevant skills that matches the job description

Do NOT use any information beyond the one provided by past experience.

Format the answer in JSON.
{format_instructions}

job description: ```{job_description}```
past experience: <<<{experience_profile}>>>
Output:
        """)
    except Exception as e:
        pass
    finally:
        # calculate elapsed time
        total_time = round((time() - start) / 60, 3)
        logger.info(f'Elapsed time to execute {cg.info()}: {total_time} min')



from enum import Enum
from pydantic import BaseModel, Field
from typing import Optional

from rqle_ai_langchain_util.utils.file_util import read_file


class ExecutionParameters(BaseModel):
    max_tokens: int = Field(description='Max number of tokens supported by LLM', default=2000)
    temperature: float = Field(description='Temperature to control the creativity of the LLM',
                               default=0.0, ge=0.0, le=1.0)
    top_p: float = Field(description='top_p controls the range of tokens considered by the LLM',
                         default=0.0, ge=0.0, le=1.0)
    presence_penalty: float = Field(description='Presence penalty aims at including a diverse range of tokens '
                                                'in the generated text', default=0.0, ge=0.0, le=1.0)
    frequency_penalty: float = Field(description='Frequency penalty aims at discouraging from repeating the '
                                                 'same words or phrases to frequently in the generated text',
                                     default=0.0, ge=0.0, le=1.0)


class PromptTypeEnum(str, Enum):
    completion = 'completion'
    chat = 'chat'
    embedding = 'embedding'


class PromptConfig(BaseModel):
    version: int = Field(description='Schema version', default=1)
    type: PromptTypeEnum = Field(description='Type of prompt', default=PromptTypeEnum.completion)
    model_name: str = Field(description='Name of the LL model')
    deployment_name: Optional[str] = Field(description='Name of the LL deployment', default='')
    parameters: ExecutionParameters = Field(description='Parameters for the execution of the LLM')

    @staticmethod
    def from_json(config_json_path: str) -> 'PromptConfig':
        """
        Create a PromptConfig from a JSON string
        :param config_json_path: Prompt configuration as a JSON string
        :return: a PromptConfig object
        """
        return PromptConfig.model_validate_json(read_file(config_json_path, 'config.json'))



import os

# define storage location of temporary audio files
AUDIO_TMP_FOLDER = os.getenv('AUDIO_TMP_FOLDER', f'{os.getcwd()}/tmp/audio')

# define storage location of the prompt config data
PROMPT_CONFIG_FOLDER = os.getenv('PROMPT_CONFIG_FOLDER', f'{os.getcwd()}/prompt_configs')

# logging configuration
LOG_LEVEL = os.getenv('LOG_LEVEL', 'DEBUG')
LOG_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
LOG_DIR = os.getenv('LOG_DIR', f'{os.getcwd()}/logs')
LOG_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'default_formatter': {
            '()': 'logging.Formatter',
            'fmt': '%(levelname)s [%(asctime)s] - %(name)s:%(funcName)s:%(lineno)d - %(message)s',
            'datefmt': LOG_DATE_FORMAT
        },
        'file_log_formatter': {
            '()': 'logging.Formatter',
            'fmt': '%(levelname)s [%(asctime)s] - %(name)s:%(lineno)d - %(message)s',
            'datefmt': LOG_DATE_FORMAT
        }
    },
    'handlers': {
        'console': {
            'formatter': 'default_formatter',
            'class': 'logging.StreamHandler',
            'level': LOG_LEVEL,
            'stream': 'ext://sys.stdout'
        },
        'file': {
            'formatter': 'file_log_formatter',
            'class': 'logging.handlers.RotatingFileHandler',
            'level': LOG_LEVEL,
            'filename': f'{LOG_DIR}/{LOG_LEVEL.lower()}.log',
            'mode': 'a',
            'encoding': 'utf-8',
            'maxBytes': 5000000,
            'backupCount': 10,
        }
    },
    'loggers': {
        '': {
            'handlers': ['console', 'file'],
            'level': LOG_LEVEL
        }
    }
}


import re

from llm_guard.input_scanners import (
    Language as InputLanguage,
    PromptInjection as InputPromptInjection,
    Toxicity as InputToxicity)
from llm_guard.input_scanners.language import MatchType as InputLanguageMatchType
from llm_guard.input_scanners.prompt_injection import MatchType as InputPromptInjectionMatchType
from llm_guard.input_scanners.toxicity import MatchType as InputToxicityMatchType
from llm_guard.output_scanners import (
    Gibberish as OutputGibberish,
    ReadingTime,
    Relevance as OutputRelevance,
    Sentiment as OutputSentiment,
    Toxicity as OutputToxicity
)
from llm_guard.output_scanners.gibberish import MatchType as OutputGibberishMatchType
from llm_guard.output_scanners.toxicity import MatchType as OutputToxicityMatchType


def validate_llm_input(input: str, languages: list[str] = None) -> (bool, str):
    '''
    :param input: the input text from the LLM
    :return: a tuple of boolean and an error message
    '''
    # check whether language is valid
    if languages is None:
        languages = ['en']
    language_scanner = InputLanguage(valid_languages=languages, match_type=InputLanguageMatchType.FULL)
    lang_sanitized_prompt, lang_is_valid, lang_risk_score = language_scanner.scan(prompt=input)
    if not lang_is_valid:
        return (False, f'Input is in a language not supported by application with {lang_risk_score} confidence.'
                       f'The list of supported languages is {languages}.')
    # check whether the input includes prompt injection
    prompt_injection_scanner = InputPromptInjection(threshold=0.5, match_type=InputPromptInjectionMatchType.FULL)
    prompt_injection_sanitized_prompt, prompt_injection_is_valid, prompt_injection_risk_score = prompt_injection_scanner.scan(prompt=input)
    if not prompt_injection_is_valid:
        return (False, f'Input contains risk of prompt injection with {prompt_injection_risk_score} confidence.')
    # check whether the input includes toxic content
    toxicity_scanner = InputToxicity(threshold=0.5, match_type=InputToxicityMatchType.SENTENCE)
    toxic_sanitized_prompt, toxic_is_valid, toxic_risk_score = toxicity_scanner.scan(prompt=input)
    if not toxic_is_valid:
        return (False, f'Input contains toxic content with {toxic_risk_score} confidence.')
    return (True, '')


def validate_llm_output(prompt: str, output: str) -> (bool, str):
    """
    :param prompt: the prompt text used by LLM
    :param output: the output text from the LLM
    :return: a tuple of boolean and an error message
    """
    # check whether the output is gibberish or nonsensical content
    gibberish_scanner = OutputGibberish(threshold=0.5, match_type=OutputGibberishMatchType.FULL)
    gibberish_sanitized_output, gibberish_is_valid, gibberish_risk_score = gibberish_scanner.scan(prompt=prompt, output=output)
    if not gibberish_is_valid:
        return (False, f'Output is gibberish with {gibberish_risk_score} confidence.')
    # check whether the output includes text with negative sentiment
    sentiment_scanner = OutputSentiment(threshold=0)
    sentiment_sanitized_output, sentiment_is_valid, sentiment_risk_score = sentiment_scanner.scan(prompt=prompt, output=output)
    if not sentiment_is_valid:
        return (False, f'Output contains content with negative sentiment with {sentiment_risk_score} confidence.')
    # check whether the output includes toxic content
    toxicity_scanner = OutputToxicity(threshold=0.5, match_type=OutputToxicityMatchType.SENTENCE)
    toxic_sanitized_output, toxic_is_valid, toxic_risk_score = toxicity_scanner.scan(prompt=prompt, output=output)
    if not toxic_is_valid:
        return (False, f'Output contains toxic content with {toxic_risk_score} confidence.')
    return (True, '')


def validate_llm_output_relevance(prompt: str, output: str) -> float:
    '''
    :param prompt: the prompt text used by LLM
    :param output: the output text from the LLM
    :return: the relevancy score of the text compaired to prompt
    '''
    # check whether the output is relevant to the prompt
    relevance_scanner = OutputRelevance(threshold=0.0)
    relevance_sanitized_output, relevance_is_valid, relevance_risk_score = relevance_scanner.scan(prompt=prompt,
                                                                                                  output=output)
    return relevance_risk_score


def validate_llm_output_reading_time(prompt: str, output: str, reading_time: int = 5) -> (bool, str):
    '''
    :param prompt: the prompt text used by LLM
    :param output: the output text from the LLM
    :param reading_time: the expected reading time in minutes
    :return: a tuple of boolean and a message
    '''
    scanner = ReadingTime(max_time=reading_time, truncate=True)
    sanitized_output, is_valid, risk_score = scanner.scan(prompt=prompt, output=output)
    if not is_valid:
        return (False, sanitized_output)
    else:
        return (True, output)


def estimate_reading_time(text: str, WPM: int = 200) -> int:
    """
    Calculate the estimated reading time for a given text based on the average words per minute.

    :param text: The input text for which the reading time needs to be estimated.
    :param WPM: The average words per minute for calculating the reading time. Default is 200.
    :return: The estimated reading time in minutes as an integer
    """
    total_words = len(re.findall(r'\w+', text))
    time_minute = total_words // WPM + 1
    if time_minute == 0:
        time_minute = 1
    return time_minute



import json

import moviepy.editor as mpe
from pydub import AudioSegment
import speech_recognition as sr

from rqle_ai_langchain_util.settings import AUDIO_TMP_FOLDER, LOG_CONFIG

# logging configuration
import logging
from logging.config import dictConfig
dictConfig(LOG_CONFIG)
logger = logging.getLogger(__name__)

SUPPORTED_SPEECH_RECOGNITION_ENGINES = ['sphinx', 'vosk']


def video_to_wav(video_path: str) -> str:
    """
    :param video_path: path to the video for which audio track needs to be extracted
    :return: path to the file containing the audio track of the video
    TODO add support for retrieving videos from YouTube
    """
    try:
        audio_filename = f'{AUDIO_TMP_FOLDER}/extracted_audio.wav'
        video = mpe.VideoFileClip(video_path)
        video.audio.write_audiofile(audio_filename)
        video.close()
        return audio_filename
    except FileNotFoundError:
        raise FileNotFoundError(f'Video file at {video_path} not found')
    except Exception as e:
        raise IOError(f'Error converting {video_path} to audio: {e}')


def chunk_audio(audio_path: str, segment_length: int, overlap: int) -> list:
    """
    :param audio_path: path to the audio file to be chunked
    :param segment_length: the length of each chunked segment
    :param overlap: the overlap between audio segments
    :return: a list of audio segment files that were generated
    """
    # initialize the variable for storing list of segment files
    tmp_segment_files = list()
    # initialize AudioSegment object based on file
    audio = AudioSegment.from_wav(audio_path)
    audio_length = len(audio)

    # determine whether audio files need to be chunked in smaller chunks
    if audio_length < segment_length:
        tmp_segment_files.append(audio_path)
    else:
        logger.info(f'Length of {audio_path} is bigger than {segment_length} and needs to be chunked.')
        # initialize the variables required for chunking
        counter = 1
        end = 0
        eof_flag = False

        # iterate through the audio file to generate sub-segments for processing
        for i in range(0, 2 * len(audio), segment_length):
            # set the indices for the first iteration
            if i == 0:
                start = 0
                end = segment_length
            else:
                start = end - overlap
                end = start + segment_length

            # check that the end index is not out of bounds
            if end > len(audio):
                end = len(audio)
                eof_flag = True

            logger.debug(f'Creating segment {counter} with start: {start} and end: {end}')
            # create a new audio segment
            segment = audio[start:end]
            # create a filename for the segment and store to tmp storage
            segment_filename = f'{AUDIO_TMP_FOLDER}/segment_{counter}.wav'
            segment.export(segment_filename, format='wav')
            tmp_segment_files.append(segment_filename)

            counter += 1

            # break if the end of the initial audio file is reached
            if eof_flag:
                break

        logger.info(f'Chunking process created {len(tmp_segment_files)} segments of audio.')

    return tmp_segment_files


def transcribe_audio(audio_path: str, segment_size: int = 60000, segment_overlap: int = 1000, language: str = 'en-US',
                     speech_recognition_engine: str = 'sphinx'):
    """
    :param audio_path: path to the audio to be transcribed
    :param segment_size: the optimal size of audio segment for processing
    :param segment_overlap: the overlap between audio segments
    :param language: the language if the audio track
    :param speech_recognition_engine: the speech recognition engine to be used for extraction text
    :return: a text version of the audio
    """
    logger.info(f'Transcribe audio from {audio_path} with segment size {segment_size} and overlap {segment_overlap}')

    # check whether the speech recognition engine for transcription is supported
    if speech_recognition_engine not in SUPPORTED_SPEECH_RECOGNITION_ENGINES:
        raise NotImplementedError(f'Speech recognition engine {speech_recognition_engine} is not supported.')

    # check whether the audio file needs to be chunked and if so, chunk it
    audio_segment_files = chunk_audio(audio_path, segment_size, segment_overlap)

    # initialize the variable for storing the transcribed text
    recognized_text = ''

    # initialize speech recognition engine
    recognizer_engine = sr.Recognizer()

    try:
        for segment_audio_file in audio_segment_files:
            logger.debug(f'Transcribing {segment_audio_file} with {speech_recognition_engine}')
            # open and read the audio segment file
            with sr.AudioFile(segment_audio_file) as audio_file:
                # read the audio track from the file
                audio_track = recognizer_engine.listen(audio_file)
                # transcribe the audio track based on the speech recognition engine
                if speech_recognition_engine == 'sphinx':
                    #TODO to be tested to determine that it works as expected
                    recognized_text += f'{json.loads(recognizer_engine.recognize_sphinx(audio_track, language=language))}\n'
                elif speech_recognition_engine == 'vosk':
                    recognized_text += f'{json.loads(recognizer_engine.recognize_vosk(audio_track, language=language))["text"]}\n'

        logger.info(f'Transcription process completed for {audio_path}.')

        # TODO create a temporary file to store the transcribed text
        # TODO create langchain document with text + metadata
        return recognized_text
    except FileNotFoundError:
        raise FileNotFoundError(f'Audio file not found at {audio_path}')
    except sr.UnknownValueError:
        raise RuntimeError(f'Audio file ({audio_path}) does not include understandable audio.')
    except Exception as e:
        raise IOError(f'Error transcribing data from {audio_path}: {e}')


if __name__ == '__main__':
    #video_to_audio('C:/Users/quent/Videos/interviews/TalkLab_Podcast_EP22_AI_Episode-1.3.mp4')
    #TODO implement support for YouTube video retrieval
    #video_to_audio('https://youtube.com/shorts/D8Gz_uOBtG0?si=Rbi2eyAB57kGx04Z', 'youtube')
    transcribed_text = transcribe_audio(audio_path=f'{AUDIO_TMP_FOLDER}/extracted_audio.wav',
                                        speech_recognition_engine='vosk')
    print(transcribed_text)



import re
from typing import List
from pydantic import BaseModel, Field

from rqle_ai_langchain_util.utils.file_util import read_file

SYSTEM_MESSAGE_PATTERN = r'(?:{{#system~}}|<<#system~>>)\n(.*?)\n(?:{{~/system}}|<<~/system>>)'
USER_MESSAGE_PATTERN = r'(?:{{#user~}}|<<#user~>>)\n(.*?)\n(?:{{~/user}}|<<~/user>>)'
ASSISTANT_MESSAGE_PATTERN = r'(?:{{#assistant~}}|<<#assistant~>>)\n(.*?)\n(?:{{~/assistant}}|<<~/assistant>>)'


class PromptTemplate(BaseModel):
    version: int = Field(description='Schema version', default=1)
    prompt: str = Field(description='Prompt', default='')
    system_message: str = Field(description='System message sets the initial context or '
                                            'instructions to the LLM on how to behave.', default='')
    user_messages: List[str] = Field(description='User message represents the input from the user.',
                                     default_factory=list)
    assistant_messages: List[str] = Field(description='Assistant message guides the output from the LLM.',
                                          default_factory=list)

    @staticmethod
    def from_text(prompt_text: str) -> 'PromptTemplate':
        """
        :param prompt_text: Prompt text
        :return: a PromptTemplate object
        """
        prompt_template = PromptTemplate()
        prompt_template.prompt = prompt_text
        # get the system message from the prompt
        system_messages = re.findall(SYSTEM_MESSAGE_PATTERN, prompt_text, re.DOTALL)
        if len(system_messages) > 0:
            prompt_template.system_message = system_messages[0]
        # get the user message from the prompt
        prompt_template.user_messages = re.findall(USER_MESSAGE_PATTERN, prompt_text, re.DOTALL)
        if len(prompt_template.user_messages) == 0:
            prompt_template.user_messages.append(prompt_text)
        # get the assistant message from the prompt
        prompt_template.assistant_messages = re.findall(ASSISTANT_MESSAGE_PATTERN, prompt_text, re.DOTALL)
        return prompt_template

    @staticmethod
    def from_json(template_json_path: str) -> 'PromptTemplate':
        """
        Create a PromptConfig from a JSON string
        :param template_json_path: Prompt configuration as a JSON string
        :return: a PromptConfig object
        """
        return PromptTemplate.model_validate_json(read_file(template_json_path, 'template.json'))



import os
from dotenv import load_dotenv

from langchain_google_genai.llms import GoogleGenerativeAI
from langchain_google_genai.chat_models import ChatGoogleGenerativeAI
from langchain_google_genai.embeddings import GoogleGenerativeAIEmbeddings

from rqle_ai_langchain_util.prompts.prompt_config import PromptConfig, PromptTypeEnum

load_dotenv()


def _load_google_gemini_llm_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain object configured for Google Gemini LLMs
    """
    llm = GoogleGenerativeAI(
        google_api_key=os.getenv('GOOGLE_API_KEY'),
        model=config.model_name,
        temperature=config.parameters.temperature,
        top_p=config.parameters.top_p,
        max_output_tokens=config.parameters.max_tokens
    )
    return llm


def _load_google_gemini_chat_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain Chat object configured for Google Gemini LLMs
    """
    llm = ChatGoogleGenerativeAI(
        google_api_key=os.getenv('GOOGLE_API_KEY'),
        model=config.model_name,
        temperature=config.parameters.temperature,
        top_p=config.parameters.top_p,
        max_output_tokens=config.parameters.max_tokens
    )
    return llm


def _load_google_gemini_embeddings_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain embeddings object configured for Google Gemini LLMs
    """
    llm = GoogleGenerativeAIEmbeddings(
        google_api_key=os.getenv('GOOGLE_API_KEY'),
        model=config.model_name
    )
    return llm


def load_google_gemini_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain object configured for Google Gemini LLMs
    """
    if config.type == PromptTypeEnum.chat:
        return _load_google_gemini_chat_from_prompt_config(config)
    elif config.type == PromptTypeEnum.completion:
        return _load_google_gemini_llm_from_prompt_config(config)
    elif config.type == PromptTypeEnum.embedding:
        return _load_google_gemini_embeddings_from_prompt_config(config)
